from __future__ import division
import numpy as np


def first(x):
    return x[0]


def second(x):
    return x[1]


def intersection_range(range1, range2):
    begin = max(range1[0], range2[0])
    end = min(range1[1], range2[1])
    if begin > end:
        return None
    return begin, end


def negate(ranges, max_len):
    end = 0
    for rbegin, rend in ranges:
        if end != rbegin:
            yield (end, rbegin)
        end = rend
    if end != max_len:
        yield (end, max_len)


def intersection(ranges1, ranges2):
    if not ranges1 or not ranges2:
        return
    it_a = iter(ranges1)
    it_b = iter(ranges2)

    ra = next(it_a)
    rb = next(it_b)

    while True:
        try:
            inter = intersection_range(ra, rb)

            if inter is not None:
                yield inter
            if second(ra) > second(rb):
                rb = next(it_b)
            else:
                ra = next(it_a)
        except StopIteration:
            break


def union(ranges1, ranges2):
    if not ranges1:
        return ranges2
    if not ranges2:
        return ranges1

    max_len = max(ranges1[-1][1], ranges2[-1][1])
    ranges1 = negate(ranges1, max_len)
    ranges2 = negate(ranges2, max_len)
    inter = (r for r in intersection(ranges1, ranges2) if r[0] != r[1])
    return negate(inter, max_len)


def len_ranges(ranges):
    return sum(
        end - begin + 1
        for begin, end in ranges
    )


def jakkard_index(ranges1, ranges2):
    intersection_len = len_ranges(intersection(ranges1, ranges2))
    union_len = len_ranges(union(ranges1, ranges2))

    if intersection_len != 0 and union_len == 0:
        print(ranges1, ranges2)
        raise Exception('union error')

    if intersection_len == 0:
        return 0
    return intersection_len / union_len


def minus(ranges1, ranges2):
    if not ranges1:
        return
    if not ranges2:
        for r in ranges1:
            yield r
        return

    max_len = max(ranges1[-1][1], ranges2[-1][1]) + 1
    ranges2 = intersection(ranges1, ranges2)
    ranges2 = negate(ranges2, max_len)
    ranges2 = [
        (begin + 1, end - 1)
        for (begin, end) in ranges2
        if end - begin >= 2
        ]
    for (begin, end) in intersection(ranges1, ranges2):
        yield (begin, end)


def true_positives_rate(test_ranges, real_ranges):
    return (
        len_ranges(intersection(test_ranges, real_ranges)) /
        len_ranges(real_ranges)
    )


def false_positives_rate(test_ranges, real_ranges, max_len=None):
    if not test_ranges:
        return 0
    return (
        len_ranges(minus(test_ranges, real_ranges)) /
        len_ranges(negate(real_ranges, max_len))
    )


def false_discovery_rate(test_ranges, real_ranges):
    test_ranges_length = len_ranges(test_ranges)
    if not test_ranges_length:
        return np.nan
    return (
        len_ranges(minus(test_ranges, real_ranges)) /
        test_ranges_length
    )


def positive_predictive_value(test_ranges, real_ranges):
    test_ranges_length = len_ranges(test_ranges)
    if not test_ranges_length:
        return np.nan

    true_positives = len_ranges(intersection(test_ranges, real_ranges))
    return true_positives / test_ranges_length


def false_negatives_rate(test_ranges, real_ranges):
    return 1 - true_positives_rate(test_ranges, real_ranges)






from __future__ import print_function, absolute_import

from .base import AbstractFinder, Phage
import math

WINDOW_SIZE = 37  # optimize this
MIN_PHAGE_IN_WINDOW = 14  # optimize this


class V1Finder(AbstractFinder):
    def __init__(self, window_size=WINDOW_SIZE, min_phage_in_window=MIN_PHAGE_IN_WINDOW):
        self.window_size = window_size
        self.min_phage_in_window = min_phage_in_window

    def find_phages(self, bacteria_npn):
        kkz, ar_positions = convert_npn(bacteria_npn)

        grad = count_grad(kkz, self.window_size)
        norm_grad = count_norm_grad(grad, self.window_size, self.min_phage_in_window)
        phages = get_phages(kkz, norm_grad)

        for phage in phages:
            # ar_here = any(
            #     phage['end'] + 2 >= pos >= phage['start'] - 2
            #     for pos in ar_positions
            # )

            yield Phage(
                begin=phage['start'],
                end=phage['end'],
                is_prophage=phage['end'] - phage['start'] < 0.8 * len(kkz),
            )


def count_grad(sequence, window_size):
    # input: 'NPPPNNPNPN??NP'-like sequence converted to [0,1,1,1,0,0,1,0,1,0,0,0,0,1], list type
    grad = []
    offset = math.floor(window_size / 2)
    for i in range(0, len(sequence)):  # for all values
        grad.append(0)
        start = int(i - offset)
        if start < 0:  # if start is negative
            start = 0
        end = int(i + offset)
        if end > len(sequence) - 1:  # if end is out of range
            end = len(sequence) - 1
        grad[i] += sum(sequence[start:end + 1])  # count sum of elements in window
    return grad


# delete all elements from grad which do not have neighbors that > minVal
def count_norm_grad(grad, window_size, min_val):
    # input: grad from previous function, window size, number of phage sequences in a window threshold
    norm_grad = list(grad)  # copy gradient
    offset = math.floor(window_size / 2)
    for i in range(0, len(norm_grad)):  # same as in previous function
        start = int(i - offset)
        if start < 0:
            start = 0
        end = int(i + offset)
        if end > len(grad) - 1:
            end = len(grad) - 1
        delete = True
        for j in range(start, end):  # check neighbors
            if norm_grad[j] > min_val:  # if at least one is bigger than minVal
                delete = False
        if delete:
            norm_grad[i] = 0
    return norm_grad


def get_phages(sequence, norm_grad):  # get coordinates of phages
    phages = []
    i = 0
    while i < len(norm_grad):
        if norm_grad[i] != 0:  # bump into not zero
            start = i
            if start < 0:
                start = 0
            end = find_end_of_range(norm_grad, start)
            if start >= end:
                break
            phage_start = find_phage_start(sequence, start)
            phage_end = find_phage_end(sequence, end)
            if phage_start < 0:
                phage_start = 0
            phages.append({"start": phage_start, "end": phage_end})
            i = end + 1
        else:
            i += 1
    return phages


def find_end_of_range(norm_grad, i_start):  # find where the sequence stops (12345321*here*0000)
    # takes normGrad and putative end position
    i = i_start
    while norm_grad[i] != 0 and i < (len(norm_grad) - 1):
        i += 1
    return i - 1


def find_phage_start(sequence, i_start):  # find first phage protein in sequence
    i = i_start
    while sequence[i] == 0 and i < (len(sequence) - 1):
        i += 1
    return i - 1


def find_phage_end(sequence, i_end):  # find last phage protein in sequence
    i = i_end
    while sequence[i] == 0:
        i -= 1
    return i + 1


def convert_npn(npn):
    # type: (str)->(list[int], list[int])

    kkz = []
    ar_positions = []

    for i, letter in enumerate(npn):
        if letter == 'P':
            kkz.append(1)
        else:
            kkz.append(0)
        if letter == 'r':
            ar_positions.append(i)
    return kkz, ar_positions





import csv
from builtins import super

from phigaro.data import read_prodigal_output, read_hmmer_output, hmmer_res_to_npn, hmmer_res_to_gc, Gene, define_taxonomy
from phigaro.finder.v2 import V2Finder
from phigaro import const
from .base import AbstractTask
from .prodigal import ProdigalTask
from .hmmer import HmmerTask
import os
from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from phigaro.to_html.preprocess import plot_html, form_sequence, if_transposable
from phigaro.to_html.html_formation import form_html_document

class RunPhigaroTask(AbstractTask):
    task_name = 'run_phigaro'

    def __init__(self, hmmer_task, prodigal_task):
        """
        :type hmmer_task: HmmerTask
        :type prodigal_task: ProdigalTask
        """
        super().__init__()
        self.hmmer_task = hmmer_task
        self.prodigal_task = prodigal_task

    def _prepare(self):
        self.finder = V2Finder(
            window_len=self.config['phigaro']['window_len'],
            threshold_min=self.config['phigaro']['threshold_min'],
            threshold_max=self.config['phigaro']['threshold_max'],
        )
        self._print_vogs = self.config['phigaro'].get('print_vogs', False)
        self._filename = self.config['phigaro'].get('filename', False)
        self._no_html = self.config['phigaro'].get('no_html', False)
        self._not_open = self.config['phigaro'].get('not_open', False)
        self._save_fasta = self.config['phigaro'].get('save_fasta', False)
        self._output = self.config['phigaro'].get('output', False)
        self._uuid = self.config['phigaro'].get('uuid', False)

    def output(self):
        return self.file('{}.tsv'.format(self.sample))

    def run(self):
        max_evalue = self.config['hmmer']['e_value_threshold']
        penalty_black = self.config['phigaro']['penalty_black']
        penalty_white = self.config['phigaro']['penalty_white']

        pvogs_black_list = const.DEFAULT_BLACK_LIST
        pvogs_white_list = const.DEFAULT_WHITE_LIST

        scaffold_set = read_prodigal_output(self.prodigal_task.output())
        hmmer_result = read_hmmer_output(self.hmmer_task.output())

        with open(self.output(), 'w') as of:
            writer = csv.writer(of, delimiter='\t')

            if self._print_vogs:
                writer.writerow(('scaffold', 'begin', 'end', 'transposable', 'taxonomy', 'vog'))
            else:
                writer.writerow(('scaffold', 'begin', 'end', 'transposable', 'taxonomy'))

            plotly_plots = []
            phage_info = []
            transposables_status = []
            phage_num = 0
            for scaffold in scaffold_set:
                phage_info.append([scaffold.name, []])
                genes = list(scaffold)  # type: list[Gene]
                npn = hmmer_res_to_npn(scaffold, hmmer_result, max_evalue=max_evalue,
                                       penalty_black = penalty_black, penalty_white = penalty_white,
                                       pvogs_black_list = pvogs_black_list, pvogs_white_list = pvogs_white_list)
                gc = hmmer_res_to_gc(scaffold, hmmer_result, max_evalue=max_evalue)

                phages = self.finder.find_phages(npn, gc)
                for phage in phages:
                    phage_num += 1
                    begin = genes[phage.begin].begin
                    end = genes[phage.end].end
                    hmmer_records = [
                        hmmer_result.min_record(hmmer_result.get_records(scaffold.name, gene.name))
                        for gene in genes[phage.begin: phage.end]
                    ]
                    hmmer_pvogs_records = (
                        record.vog_name
                        for record in hmmer_records
                        if record and record.evalue <= max_evalue
                    )
                    pvogs_records_str = ', '.join(hmmer_pvogs_records)
                    taxonomy = define_taxonomy(pvogs_records_str)
                    hmmer_records = [
                        record
                        for record in hmmer_records
                        if record and record.evalue <= max_evalue
                    ]
                    transposable = if_transposable(hmmer_records)
                    if self._print_vogs:
                        writer.writerow((scaffold.name, begin, end, transposable, taxonomy,  pvogs_records_str))
                    else:
                        writer.writerow((scaffold.name, begin, end, transposable, taxonomy))

                    sequence, record = form_sequence(self._filename, '%s_prophage_%d' % (scaffold.name, phage_num), begin, end, scaffold.name)
                    transposables_status.append(transposable)
                    the_phage_info = [begin, end, taxonomy, sequence, pvogs_records_str]
                    phage_info[-1][1].append(the_phage_info)
                    if self._save_fasta:
                        with open(self._output+'.fasta', 'a') as f:
                            SeqIO.write(SeqRecord(record, id='%s_prophage_%d' % (scaffold.name, phage_num),
                                                  description='%s_prophage_%d' % (scaffold.name, phage_num)), f, "fasta")
                    if (not self._no_html) and (self._output):
                        plotly_plots.append(plot_html(hmmer_records, begin, end))
                phage_info = phage_info if len(phage_info[-1][1]) > 0 else phage_info[:-1]
            if (len(phage_info) > 0) and (not self._no_html) and (self._output):
                html = form_html_document(phage_info, transposables_status, plotly_plots, self._filename, self._uuid)
                with open(self._output+'.html', 'w') as f:
                    f.write(html)
                if not self._not_open:
                    os.system('xdg-open '+self._output+'.html'+' > /dev/null 2>/dev/null')


import abc
from os import unlink
from os.path import exists

import six

from phigaro.context import Context
from phigaro.batch.task.path import directory, file


@six.add_metaclass(abc.ABCMeta)
class AbstractTask(object):
    task_name = None

    def __init__(self):
        if self.task_name is None:
            raise Exception("{}.task_name must by set".format(self.__class__.__name__))

        self.directory()
        self.context = Context.instance()
        self.sample = self.context.sample
        self.config = self.context.config
        self._prepare()

    def _prepare(self):
        pass

    @abc.abstractmethod
    def output(self):
        """
        :rtype: str
        """
        pass

    @abc.abstractmethod
    def run(self):
        pass

    def directory(self, *items):
        return directory(self.task_name, *items)

    def file(self, *items):
        return file(self.task_name, *items)

    def clean(self):
        out = self.output()
        if exists(out):
            unlink(out)








from builtins import super

import os

from .base import AbstractTask
import sh


class GeneMarkTask(AbstractTask):
    task_name = 'gene_mark'

    def __init__(self, input):
        super().__init__()

        self.input = input
        self._lst_file = input + '.lst'
        self.gene_mark = (
            sh.Command(self.config['genemark']['bin'])
            .bake(
                m=self.config['genemark']['mod_path'],
                A=self.output(),
            )
        )

    def output(self):
        return self.file('{}.faa'.format(self.sample))

    def run(self):
        self.gene_mark(self.input)

    def clean(self):
        os.unlink(self._lst_file)



from __future__ import absolute_import

import argparse
import logging
import multiprocessing
import os
import sys
import uuid
from os.path import join, exists

import yaml

from phigaro.context import Context
from phigaro.batch.runner import run_tasks_chain
from phigaro.batch.task.path import sample_name
from phigaro.batch.task.prodigal import ProdigalTask
from phigaro.batch.task.hmmer import HmmerTask
from phigaro.batch.task.dummy import DummyTask
from phigaro.batch.task.preprocess import PreprocessTask
from phigaro.batch.task.run_phigaro import RunPhigaroTask
from phigaro._version import __version__

def parse_substitute_output(subs):
    subs = subs or []
    res = {}
    for sub in subs:
        task_name, output = sub.split(":")
        res[task_name] = DummyTask(output, task_name)
    return res


def create_task(substitutions, task_class, *args, **kwargs):
    # TODO: refactor to class Application
    task = task_class(*args, **kwargs)
    if task.task_name in substitutions:
        print('Substituting output for {}: {}'.format(
            task.task_name, substitutions[task.task_name].output()
        ))

        return substitutions[task.task_name]
    return task

def clean_fold():
    is_empty = True
    for root, dirs, files in os.walk('proc', topdown=False):
        for name in files:
            is_empty = False
            break
        if is_empty:
            for name in dirs:
                os.rmdir(os.path.join(root, name))
    if is_empty:
        os.rmdir('proc')


def main():
    default_config_path = join(os.getenv('HOME'), '.phigaro', 'config.yml')
    parser = argparse.ArgumentParser(
        prog='phigaro',
        description='Phigaro is a scalable command-line tool for predictions phages and prophages '
                    'from nucleid acid sequences',
    )

    parser.add_argument('-V', '--version', action='version',
                        version='%(prog)s {version}'.format(version=__version__))
    parser.add_argument('-f', '--fasta-file', help='Assembly scaffolds/contigs or full genomes, required',
                        required=True)
    parser.add_argument('-c', '--config', default=default_config_path, help='Config file, not required')
    parser.add_argument('-v', '--verbose', action='store_true', help=argparse.SUPPRESS)
    parser.add_argument('-p', '--print-vogs', help='Print phage vogs for each region', action='store_true')
    parser.add_argument('-e', '--extension', default=['html'], nargs='+', help='Type of the output: html, txt or stdout. Default is html. You can specify several file formats with a space as a separator. Example: -e txt html stdout.')
    parser.add_argument('-o', '--output', default=False, help='Output filename for html and txt outputs. Required by default, but not required for stdout only output.')
    parser.add_argument('--not-open', help='Do not open html file automatically, if html output type is specified.', action='store_true')
    parser.add_argument('-t', '--threads',
                        type=int,
                        default=multiprocessing.cpu_count(),
                        help='Num of threads ('
                             'default is num of CPUs={})'.format(multiprocessing.cpu_count()))
    parser.add_argument('--no-cleanup', action='store_true', help=argparse.SUPPRESS)
    parser.add_argument('-S', '--substitute-output', action='append', help=argparse.SUPPRESS)

    args = parser.parse_args()

    logging.basicConfig(level=logging.INFO if args.verbose else logging.WARN)
    logging.getLogger('sh.command').setLevel(logging.WARN)

    logger = logging.getLogger(__name__)

    if not exists(args.config):
        # TODO: pretty message
        print('Please, create config file using phigaro-setup script')
        exit(1)

    args.extension = [atype.lower() for atype in args.extension]
    if (not args.output) and (args.extension != ['stdout']):
        print('Error! Argument -o/--output is required or change the type of the output to stdout.')
        exit(1)

    with open(args.config) as f:
        logger.info('Using config file: {}'.format(args.config))
        config = yaml.load(f, Loader=yaml.FullLoader)

    config['phigaro']['print_vogs'] = args.print_vogs
    config['phigaro']['filename'] = args.fasta_file
    config['phigaro']['no_html'] = True if 'html' not in args.extension else False
    config['phigaro']['not_open'] = args.not_open
    config['phigaro']['output'] = args.output
    config['phigaro']['uuid'] = uuid.uuid4().hex

    filename = args.fasta_file
    sample = '{}-{}'.format(
        sample_name(filename),
        config['phigaro']['uuid']
    )

    Context.initialize(
        sample=sample,
        config=config,
        threads=args.threads,
    )

    substitutions = parse_substitute_output(args.substitute_output)

    preprocess_task = create_task(substitutions,
                       PreprocessTask,
                       filename)

    prodigal_task = create_task(substitutions,
                                ProdigalTask,
                                preprocess_task=preprocess_task)
    hmmer_task = create_task(substitutions,
                             HmmerTask,
                             prodigal_task=prodigal_task)

    run_phigaro_task = create_task(substitutions,
                                   RunPhigaroTask,
                                   prodigal_task=prodigal_task,
                                   hmmer_task=hmmer_task)

    tasks = [
        preprocess_task,
        prodigal_task,
        hmmer_task,
        run_phigaro_task
    ]
    task_output_file = run_tasks_chain(tasks)

    if ('txt' in args.extension) or ('stdout' in args.extension):
        with open(task_output_file) as f:
            if 'txt' in args.extension:
                out_f = open(args.output, 'w')
                for line in f:
                    out_f.write(line)
            if 'stdout' in args.extension:
                out_f = sys.stdout
                for line in f:
                    out_f.write(line)
                out_f.close()

    if not args.no_cleanup:
        for t in tasks:
            t.clean()
        clean_fold()

if __name__ == '__main__':
    main()

